/* functions from structure.js */
function get_depth(orig_node, probe_node, nodes){
    if (typeof nodes[probe_node] != 'undefined'){
        if (nodes[probe_node].parent != "n0"){
            parent_id = nodes[probe_node].parent;
            if (typeof nodes[parent_id] != 'undefined'){
                if (nodes[parent_id].kind != "edu" && (nodes[probe_node].relname == "span" || nodes[parent_id].kind == "multinuc" && nodes[probe_node].reltype == "multinuc")){
                    nodes[orig_node].depth += 1;
                }
            }
            get_depth(orig_node, parent_id, nodes);
        }
    }
}

function get_left_right(node_id, nodes, min_left, max_right){
	if (nodes[node_id].parent != "n0" && node_id != "n0" && typeof nodes[nodes[node_id].parent] != 'undefined'){
		parent = nodes[nodes[node_id].parent];
		if (min_left > nodes[node_id].left || min_left == 0){
			if (nodes[node_id].left != 0){
				min_left = nodes[node_id].left;
			}
		}
		if (max_right < nodes[node_id].right || max_right == 0){
			max_right = nodes[node_id].right;
		}
		if (nodes[node_id].relname == "span" && parent.kind =="span"){
			if (parent.left > min_left || parent.left == 0){
				parent.left = min_left;
			}
			if (parent.right < max_right){
				parent.right = max_right;
			}
		}
		else if (nodes[node_id].relname != "span"){
			if (parent.kind == "multinuc" && get_rel_type(nodes[node_id].relname) =="multinuc"){
				if (parent.left > min_left || parent.left == 0){
					parent.left = min_left;
				}
				if (parent.right < max_right){
					parent.right = max_right;
				}
			}
		}
		get_left_right(parent.id, nodes, min_left, max_right);
	}
}

/* new functions */

function build_canvas(rels, nodes){
	current_doc = "none";
	current_project= "none";
	show_signals = false;
	/* GRAPHICAL PARAMETERS */
	top_spacing = 20;
	layer_spacing = 60;

	signal_data = null;
	if (signal_data == null){
		signal_data = [];
	}

	validations="validate_empty;validate_flat;validate_mononuc";
	if (validations == null){
		//validations = get_project_validations(current_project);
		//NOT IMPLEMENTED
	}

	cpout = '<div id="container" class="container">\n';
	if (show_signals){
		cpout += `	            <button id="show-all-signals">
	              Show All Signal Tokens
	            </button>\n`;

	cpout += `	            <div class="signal-drawer">
	              <div id="signal-list"> </div>

				  <div class="signal-drawer__row">
				    <button id="new-signal" class="signal-drawer__create-new-button">
	  			      <i class="fa fa-plus" title="New Signal"> </i>
				      New Signal
				    </button>
				  </div>
				  <div class="signal-drawer__row" style="text-align: center;padding-top:20px;">
	  		        <button id="save-signals" class="signal-drawer__save-button">
	  		          <i class="fa fa-check"> </i>
	  		          Save Changes
	  		        </button>
	  		        <button id="cancel-signals" class="signal-drawer__cancel-button">
	  		          <i class="fa fa-ban"> </i>
	  		          Cancel
	  		        </button>
	  		      </div>
	            </div>
	    `;
	}

	cpout += '''<div id="canvas" class="canvas">'''
	cpout += '\t<p id="document_name">Document: <b>' + current_doc + '</b> (project: <i>' + current_project + '</i>)</p>'
	cpout += '''<div id="inner_canvas">'''
	cpout += '<script src="./script/structure.js"></script>'

	multi_options = ""
	rst_options = ""
	rel_list = []

	rel_kinds = {}
	if def_rstrel == "":
		def_rstrel = sorted([rel[0] for rel in rels if rel[1] != "multinuc"])[0]  # Take alphabetically first rel
	if def_rstrel == "":
		def_rstrel = "elaboration_r"
	if def_multirel == "":
		def_multirel = sorted([rel[0] for rel in rels if rel[1] == "multinuc"])[0]  # Take alphabetically first rel
	if def_multirel == "":
		def_multirel = "joint_m"
	for rel in rels:
		rel_list.append(rel[0])
		if rel[1] == "multinuc":
			multi_options += "<option value='" + rel[0] + "'>" + rel[0].replace("_m", "") + '</option>\n'
			rel_kinds[rel[0]] = "multinuc"
		else:
			rst_options += "<option value='" + rel[0] + "'>" + rel[0].replace("_r", "") + '</option>\n'
			rel_kinds[rel[0]] = "rst"
	rel_list.sort()
	multi_options += "<option value='" + def_rstrel + "'>(satellite...)</option>\n"

	for key in nodes:
		node = nodes[key]
		get_depth(node, node, nodes, doc=current_doc, project=current_project, user=user)

	for key in nodes:
		if nodes[key].kind == "edu":
			get_left_right(key, nodes, 0, 0, rel_kinds)

	signals = {}

	for signal in signal_data:
		s_id, s_type, subtype, tokens = signal
		if s_id not in signals:
			signals[s_id] = list()
		if tokens:
			tokens = list(map(int, tokens.split(",")))
		else:
			tokens = []
		signals[s_id].append({'type': s_type,
							  'subtype': subtype,
							  'tokens': tokens})
	cpout += '<script>'
	cpout += 'window.rstWebSignals = ' + json.dumps(signals) + ';'
	cpout += 'window.rstWebSignalTypes = ' + json.dumps(get_signal_types_dict(current_doc, current_project),
														sort_keys=True) + ';'
	cpout += 'window.rstWebDefaultSignalType = Object.keys(window.rstWebSignalTypes)[0];'
	cpout += 'if (window.rstWebDefaultSignalType != null) {window.rstWebDefaultSignalSubtype = window.rstWebSignalTypes[window.rstWebDefaultSignalType][0];}'
	cpout += '</script>'

	anchors = {}
	pix_anchors = {}

	#dbpath = os.path.dirname(os.path.realpath(__file__)) + os.sep + "rstweb.db"
	#conn = sqlite3.connect(dbpath)
	#cur = conn.cursor()

	# Calculate anchor points for nodes
	# First get proportional position for anchor

	#nodes_data = get_multinuc_nodes_data(cur, current_doc, current_project, user)
	#id, parent, left, right
	nodes_data = [[n.id, n.parent, n.left, n.right] for n in itervalues(nodes) if n.relkind == "multinuc"]

	#doc_relname_to_reltype = get_doc_relname_to_reltype(cur, current_doc, current_project)
	doc_relname_to_reltype = {"span": "span", "": "span"}
	doc_relname_to_reltype.update(rel_kinds)

	node_to_lr = {}

	for node in nodes_data:
		if not node[1] in node_to_lr:
			node_to_lr[node[1]] = {
				'left': {},
				'right': {}
			}
		if not node[2] in node_to_lr[node[1]]['left']:
			node_to_lr[node[1]]['left'][node[2]] = node[0]
		if not node[3] in node_to_lr[node[1]]['right']:
			node_to_lr[node[1]]['right'][node[3]] = node[0]

	for key in sorted(nodes, key=lambda id: nodes[id].depth, reverse=True):
		node = nodes[key]
		if node.kind == "edu":
			anchors[node.id] = "0.5"
		if node.parent != "0":
			parent = nodes[node.parent]
			parent_wid = (parent.right - parent.left + 1) * 100 - 4
			child_wid = (node.right - node.left + 1) * 100 - 4
			if node.relname == "span":
				if node.id in anchors:
					anchors[parent.id] = str(
						((node.left - parent.left) * 100) / parent_wid + float(anchors[node.id]) * float(
							child_wid / parent_wid))
				else:
					anchors[parent.id] = str(
						((node.left - parent.left) * 100) / parent_wid + (0.5 * child_wid) / parent_wid)
			elif node.relkind == "multinuc" and parent.kind == "multinuc":
				# For multinucs, the anchor is in the middle between leftmost and rightmost of the multinuc children
				# (not including other rst children)
				node_lr = node_to_lr[node.parent]
				lr = [min(node_lr['left'].keys()), max(node_lr['right'].keys())]

				lr_wid = (lr[0] + lr[1]) / 2
				lr_ids = (node_to_lr[node.parent]['left'][lr[0]], node_to_lr[node.parent]['right'][lr[1]])

				left_child = lr_ids[0]
				right_child = lr_ids[1]
				if left_child == right_child:
					anchors[parent.id] = "0.5"
				else:
					if left_child in anchors and right_child in anchors:  # both leftmost and rightmost multinuc children have been found
						len_left = nodes[left_child].right - nodes[left_child].left + 1
						len_right = nodes[right_child].right - nodes[right_child].left + 1
						anchors[parent.id] = str(((float(anchors[left_child]) * len_left * 100 + float(
							anchors[right_child]) * len_right * 100 + (nodes[
																		   right_child].left - parent.left) * 100) / 2) / parent_wid)
					else:
						anchors[parent.id] = str((lr_wid - parent.left + 1) / (parent.right - parent.left + 1))

			else:
				if not parent.id in anchors:
					anchors[parent.id] = "0.5"

	# Place anchor element to center on proportional position relative to parent
	for key in nodes:
		node = nodes[key]
		pix_anchors[node.id] = str(
			int(3 + node.left * 100 - 100 - 39 + float(anchors[node.id]) * ((node.right - node.left + 1) * 100 - 4)))

	# Check that span and multinuc buttons should be used (if the interface is not used for RST, they may be disabled)
	if int(get_schema()) > 2:
		use_span_buttons = True if get_setting("use_span_buttons") == "True" else False
		use_multinuc_buttons = True if get_setting("use_multinuc_buttons") == "True" else False
	else:
		use_span_buttons = True
		use_multinuc_buttons = True

	if sys.version_info[0] == 2:
		lambda_button = "Λ".decode("utf-8")
	else:
		lambda_button = "Λ"

	tok_count = 0
	for key in sorted(nodes.keys(), key=int):
		node = nodes[key]
		if node.kind != "edu":
			g_wid = str(int((node.right - node.left + 1) * 100 - 4))
			cpout += '<div id="lg' + node.id + '" class="group" style="left: ' + str(
				int(node.left * 100 - 100)) + '; width: ' + g_wid + '; top:' + str(int(
				top_spacing + layer_spacing + node.depth * layer_spacing)) + 'px; z-index:1"><div id="wsk' + node.id + '" class="whisker" style="width:' + g_wid + ';"></div></div>'
			cpout += '<div id="g' + node.id + '" class="num_cont" style="position: absolute; left:' + pix_anchors[
				node.id] + 'px; top:' + str(
				int(4 + top_spacing + layer_spacing + node.depth * layer_spacing)) + 'px; z-index:' + str(int(200 - (
						node.right - node.left))) + '"><table class="btn_tb"><tr><td rowspan="2"><button id="unlink_' + node.id + '"  title="unlink this node" class="minibtn" onclick="act(' + "'up:" + node.id + ",0'" + ');">X</button></td><td rowspan="2"><span class="num_id">' + str(
				int(node.left)) + "-" + str(int(node.right)) + '</span></td>'
			if use_span_buttons:
				cpout += '<td><button id="aspan_' + node.id + '" title="add span above" class="minibtn" onclick="act(' + "'sp:" + node.id + "'" + ');">T</button></td>'
			cpout += '</tr>'
			if use_multinuc_buttons:
				cpout += '<tr><td><button id="amulti_' + node.id + '" title="add multinuc above" class="minibtn" onclick="act(' + "'mn:" + node.id + "'" + ');">' + lambda_button + '</button></td></tr>'
			cpout += '</table></div><br/>'

		elif node.kind == "edu":
			cpout += '<div id="edu' + str(node.id) + '" class="edu" title="' + str(node.id) + '" style="left:' + str(
				int(node.id) * 100 - 100) + '; top:' + str(
				int(top_spacing + layer_spacing + node.depth * layer_spacing)) + '; width: 96px">'
			cpout += '<div id="wsk' + node.id + '" class="whisker" style="width:96px;"></div><div class="edu_num_cont"><table class="btn_tb"><tr><td rowspan="2"><button id="unlink_' + node.id + '" title="unlink this node" class="minibtn" onclick="act(' + "'up:" + node.id + ",0'" + ');">X</button></td><td rowspan="2"><span class="num_id">&nbsp;' + str(
				int(node.left)) + '&nbsp;</span></td>'
			if use_span_buttons:
				cpout += '<td><button id="aspan_' + node.id + '" title="add span above" class="minibtn" onclick="act(' + "'sp:" + node.id + "'" + ');">T</button></td>'
			cpout += '</tr>'
			if use_multinuc_buttons:
				cpout += '<tr><td><button id="amulti_' + node.id + '" title="add multinuc above" class="minibtn" onclick="act(' + "'mn:" + node.id + "'" + ');">' + lambda_button + '</button></td></tr>'
			cpout += '</table></div>'

			for tok in node.text.split(" "):
				tok_count += 1
				cpout += '<span id="tok' + str(tok_count) + '" class="tok">' + tok + '</span> '

			cpout += '</div>'

	# Serialize data in hidden input for JavaScript
	cpout += '<input id="data" name="data" type="hidden" '
	hidden_val = ""

	#doc_relname_to_reltype = get_doc_relname_to_reltype(cur, current_doc, current_project)
	for key in nodes:
		node = nodes[key]
		reltype = doc_relname_to_reltype.get(node.relname)

		if node.relname:
			safe_relname = node.relname
		else:
			safe_relname = "none"
		if node.kind == "edu":
			hidden_val += "n" + node.id + ",n" + node.parent + ",e," + str(
				int(node.left)) + "," + safe_relname + "," + reltype + ";"
		elif node.kind == "span":
			hidden_val += "n" + node.id + ",n" + node.parent + ",s,0," + safe_relname + "," + reltype + ";"
		else:
			hidden_val += "n" + node.id + ",n" + node.parent + ",m,0," + safe_relname + "," + reltype + ";"
	hidden_val = hidden_val[0:len(hidden_val) - 1]
	cpout += 'value="' + hidden_val + '"/>'

	cpout += '<input id="def_multi_rel" type="hidden" value="' + def_multirel + '"/>\n'
	cpout += '<input id="def_rst_rel" type="hidden" value="' + def_rstrel + '"/>\n'
	cpout += '<input id="undo_log" type="hidden" value=""/>\n'
	cpout += '<input id="redo_log" type="hidden" value=""/>\n'
	cpout += '<input id="undo_state" type="hidden" value=""/>\n'
	cpout += '<input id="logging" type="hidden" value=""/>\n'
	cpout += '<input id="validations" type="hidden" value="' + validations + '"/>\n'
	cpout += '<input id="use_span_buttons" type="hidden" value="' + str(use_span_buttons) + '"/>\n'
	cpout += '<input id="use_multinuc_buttons" type="hidden" value="' + str(use_multinuc_buttons) + '"/>\n'

	cpout += '''	<script src="./script/jquery.jsPlumb-1.7.5-min.js"></script>

				<script>
				'''

	cpout += 'function list_rels(){return "' + ",".join(rel_list) + '";}'

	cpout += 'function select_my_rel(options,my_rel){'
	cpout += 'var multi_options = `' + multi_options + '`;\n'
	cpout += 'var rst_options = `' + rst_options + '`;\n'
	cpout += 'if (options =="multi"){options = multi_options;} else {options=rst_options;}\n'
	cpout += '		replaced = options.replace("<option value=' + "'" + '"' + '+my_rel+' + '"' + "'" + '","<option selected=' + "'" + 'selected' + "'" + ' value=' + "'" + '"' + '+my_rel+' + '"' + "'" + '");\n'
	cpout += '      const longrel = /([a-z]{4})[a-z]+-([a-z]+<.option)/gi;\n'
	cpout += '      replaced = replaced.replace(longrel,"$1-$2"); //shorten long names\n'
	cpout += '      return replaced;\n'
	cpout += '			}\n'
	cpout += '''function make_relchooser(id,option_type,rel){
		    var s = "<div id='seldiv"+id.replace("n","")+"' style='white-space:nowrap;'>";
		    s += make_signal_button(id);
		    s += "<select id='sel"+id.replace("n","")+"' class='rst_rel' onchange='crel(" + id.replace("n","") + ",this.options[this.selectedIndex].value);'>" + select_my_rel(option_type,rel) + "</select>";
		    return $(s);
		}'''
	# todo: make a flag that controls whether signals are on
	cpout += 'var signalsEnabled = ' + ('true;' if get_setting("signals") == "True" else 'false;')
	cpout += '''function make_signal_button(id) {
			if (signalsEnabled) {
				var text = window.rstWebSignals[id]
						? window.rstWebSignals[id].length
						: "S";
				var classes = window.rstWebSignals[id] && window.rstWebSignals[id].length > 0
						? "minibtn minibtn--with-signals"
						: "minibtn";
				return '<button title="add signals" class="' + classes + '" onclick="open_signal_drawer(\\'' + id.replace('n','') + '\\')">' + text + '</button>';
			} else {
				return '';
			}
		}'''
	cpout += '''
				jsPlumb.importDefaults({
				PaintStyle : {
					lineWidth:2,
					strokeStyle: 'rgba(0,0,0,0.5)'
				},
				Endpoints : [ [ "Dot", { radius:1 } ], [ "Dot", { radius:1 } ] ],
				  EndpointStyles : [{ fillStyle:"#000000" }, { fillStyle:"#000000" }],
				  Anchor:"Top",
		            Connector : [ "Bezier", { curviness:50 } ]
				})
				 jsPlumb.ready(function() {

		jsPlumb.setContainer(document.getElementById("inner_canvas"));
		'''

	cpout += "jsPlumb.setSuspendDrawing(true);"

	for key in nodes:
		node = nodes[key]
		if node.kind == "edu":
			node_id_str = "edu" + node.id
		else:
			node_id_str = "g" + node.id
		cpout += 'jsPlumb.makeSource("' + node_id_str + '", {anchor: "Top", filter: ".num_id", allowLoopback:false});'
		cpout += 'jsPlumb.makeTarget("' + node_id_str + '", {anchor: "Top", filter: ".num_id", allowLoopback:false});'

	# Connect nodes
	for key in nodes:
		node = nodes[key]
		if node.parent != "0":
			parent = nodes[node.parent]
			if node.kind == "edu":
				node_id_str = "edu" + node.id
			else:
				node_id_str = "g" + node.id
			if parent.kind == "edu":
				parent_id_str = "edu" + parent.id
			else:
				parent_id_str = "g" + parent.id

			if node.relname == "span":
				cpout += 'jsPlumb.connect({source:"' + node_id_str + '",target:"' + parent_id_str + '", connector:"Straight", anchors: ["Top","Bottom"]});'
			elif parent.kind == "multinuc" and node.relkind == "multinuc":
				cpout += 'jsPlumb.connect({source:"' + node_id_str + '",target:"' + parent_id_str + '", connector:"Straight", anchors: ["Top","Bottom"], overlays: [ ["Custom", {create:function(component) {return make_relchooser("' + node.id + '","multi","' + node.relname + '");},location:0.2,id:"customOverlay"}]]});'
			else:
				cpout += 'jsPlumb.connect({source:"' + node_id_str + '",target:"' + parent_id_str + '", overlays: [ ["Arrow" , { width:12, length:12, location:0.95 }],["Custom", {create:function(component) {return make_relchooser("' + node.id + '","rst","' + node.relname + '");},location:0.1,id:"customOverlay"}]]});'

	cpout += '''

			jsPlumb.setSuspendDrawing(false,true);


			jsPlumb.bind("connection", function(info) {
			   source = info.sourceId.replace(/edu|g/,"")
			   target = info.targetId.replace(/edu|g/g,"")
			});

			jsPlumb.bind("beforeDrop", function(info) {
			    $(".minibtn").prop("disabled",true);

		'''

	cpout += '''
				var node_id = "n"+info.sourceId.replace(/edu|g|lg/,"");
				var new_parent_id = "n"+info.targetId.replace(/edu|g|lg/,"");

				nodes = parse_data();
				new_parent = nodes[new_parent_id];
				relname = nodes[node_id].relname;
				new_parent_kind = new_parent.kind;
				if (nodes[node_id].parent != "n0"){
					old_parent_kind = nodes[nodes[node_id].parent].kind;
				}
				else
				{
					old_parent_kind ="none";
				}

				if (info.sourceId != info.targetId){
					if (!(is_ancestor(new_parent_id,node_id))){
						jsPlumb.select({source:info.sourceId}).detach();
						if (new_parent_kind == "multinuc"){
							relname = get_multirel(new_parent_id,node_id,nodes);
							jsPlumb.connect({source:info.sourceId, target:info.targetId, connector:"Straight", anchors: ["Top","Bottom"], overlays: [ ["Custom", {create:function(component) {return make_relchooser(node_id,"multi",relname);},location:0.2,id:"customOverlay"}]]});
						}
						else{
							jsPlumb.connect({source:info.sourceId, target:info.targetId, overlays: [ ["Arrow" , { width:12, length:12, location:0.95 }],["Custom", {create:function(component) {return make_relchooser(node_id,"rst",relname);},location:0.1,id:"customOverlay"}]]});
						}
						new_rel = document.getElementById("sel"+ node_id.replace("n","")).value;
					    act('up:' + node_id.replace("n","") + ',' + new_parent_id.replace("n",""));
						update_rel(node_id,new_rel,nodes);
						recalculate_depth(parse_data());
					}
				}

			    $(".minibtn").prop("disabled",false);

			});

		});
			nodes = parse_data();
			show_warnings(nodes);

				</script>

	'''

	cpout += '''
				</div>
				</div>
				</div>
				<div id="anim_catch" class="anim_catch">&nbsp;</div>
		'''
	return cpout;
}
